/* Copyright (C) 2007 GFRN systems

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The latest version of this program may be found at
   http://CQiNet.sourceforge.net

   $Log: nodes.cpp,v $
   Revision 1.8  2008/04/25 05:24:59  wb6ymh
   Removed initialization of DisableRateConversion for Nodes constructor.

   Revision 1.7  2008/04/25 05:05:49  wb6ymh
   1. Added support for command prefix decoding.
   2. Added numerious checks to make sure the current port still exists before
      trying to use it.
   3. Added code to DecodeDTMF to set CurrentPort so the TLB_PORT environment
      variable gets set to the port issuing the command.

   Revision 1.6  2008/04/11 14:12:03  wb6ymh
   Removed .dtmfgen command, dtmf generation is now handled by the .tonegen
   command.

   Revision 1.5  2008/03/18 20:03:40  wb6ymh
   Modified code to use bCwIsID and bAnnoucementIsID to track which 
   transmissions should set LastActivity.  Previously LastActivity wasn't set 
   by anything other than an announcement which prevented IDs from running 
   when they should.

   Revision 1.4  2008/03/12 21:43:05  wb6ymh
   Added code to free NodeName to Node::~Node.

   Revision 1.3  2008/03/11 14:06:30  wb6ymh
   Modified SendCTone to check LNK_FLG_CTONE (set by CToneMap)  before sending
   Ctones to a node.

   Revision 1.2  2008/03/09 18:19:47  wb6ymh
   Added headers needed by FreeBSD.

   Revision 1.1  2008/03/09 17:10:13  wb6ymh
   Inital import - files broken out of old audio.cpp and audio.h due to
   major restructuring of the internal infrastructure to treat VoIP connections
   as pseudo ports.

*/ 
#include "common.h"

// FreeBSD, Linux, etc..
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
   #include <sys/time.h>
   #include <time.h>
#else
   #ifdef HAVE_SYS_TIME_H
      #include <sys/time.h>
   #else
      #include <time.h>
   #endif
#endif
#include <unistd.h>
#include <errno.h>
#include <string.h>
#ifdef HAVE_FCNTL_H
   #include <fcntl.h>
#endif
#include <netdb.h>
#include <ctype.h>
#ifdef HAVE_SYS_TIMEB_H
   #include <sys/timeb.h>
#endif
#include <math.h>
#include <typeinfo>

#ifdef HAVE_SYS_SYSLIMITS_H
#include <sys/syslimits.h> // for PATH_MAX
#endif

#include <sys/file.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include "avl.h"
#include "main.h"
#include "configvars.h"
#include "configfile.h"
#include "conference.h"
#include "linkbox.h"
#include "eventhook.h"
#include "tbd.h"

#include "nodes.h"
#include "ports.h"
#include "voip.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif

#define TTS_OUT_FILENAME   "tts.wav"
#define TTS_IN_FILENAME    "tts.in"

#define MAX_MIXED_LEVEL 32000

/*
bytes      variable      description
0  - 3     'RIFF'/'RIFX' Little/Big-endian
4  - 7     wRiffLength   length of file minus the 8 byte riff header
8  - 11    'WAVE'
12 - 15    'fmt '
16 - 19    wFmtSize       length of format chunk minus 8 byte header 
20 - 21    wFormatTag     identifies PCM, ULAW etc
22 - 23    wChannels      
24 - 27    dwSamplesPerSecond  samples per second per channel
28 - 31    dwAvgBytesPerSec    non-trivial for compressed formats
32 - 33    wBlockAlign         basic block size
34 - 35    wBitsPerSample      non-trivial for compressed formats
*/

typedef struct {
   char  Riff[4];             // 'RIFF'/'RIFX' Little/Big-endian
   int   wRiffLength;         // length of file minus the 8 byte riff header
   char  Wave[4];             // 'WAVE'
   char  Fmt[4];              // 'fmt '
   int   wFmtSize;            // length of format chunk minus 8 byte header 
   short wFormatTag;          // identifies PCM, ULAW etc
   short wChannels;
   int   dwSamplesPerSecond;  // samples per second per channel
   int   dwAvgBytesPerSec;    // non-trivial for compressed formats
   short wBlockAlign;         // basic block size
   short wBitsPerSample;      // non-trivial for compressed formats
} WaveHdr;

struct avl_table *NodeTree;   // sorted by node name
int bTTS_Running = FALSE;
int MixBuffer[AUDIO_BUF_SIZE];
int16 ToneBuf[AUDIO_BUF_SIZE];

int SayConnectResult(ClientInfo *p,ConfClient *pCC,char *Arg);
int SayDisconnectResult(ClientInfo *p,ConfClient *pCC,char *Arg);
void AnnounceConnect(char *Args);
void AnnounceDisconnect(char *Args);

struct SayResultHandler {
   const char *Cmd;
   int (*Handler)(ClientInfo *p,ConfClient *pCC,char *Arg);
} SayResultHandlers[] = {
   {"connect", SayConnectResult},
   {"disconnect", SayDisconnectResult},
   { NULL}  // end of table
};

char *StdTbdCommandText[] = {
   // command not found or ambiguous
   "unknown command",      
   
   // command requires disk and disk is disabled
   "no disk is available"  
   
   // Specified station not found
   "station not found",    
   
   // no info available for station (.info cmd)
   "no information on station is available", 

   // invalid command argument count
   "error, invalid number command arguments",

   // Already connected (.connect cmd)
   "you are already connected to that station",

   // no one is talking (.mute cmd)
   "no one is talking",

   // invalid command argument
   "command argument is not valid",

   // error opening a file
   "error can not open file",

   // data is chat text
   "",

   // timeout waiting for response from tbd ... it's probably not running
   "error timeout",

   // last "command" was sent as chat text
   "",

   // Specified port not found
   "port not found",
   
   // Say command has completed
   ""
};

/* A -> Z */
char *CwAlphaLookup[] = {
   ".-","-...","-.-.","-..",".","..-.","--.","....","..",".---","-.-",
   ".-..","--","-.","---",".--.","--.-",".-.","...","-","..-","...-",
   ".--","-..-","-.--","--.."
};

/* 0 -> 9 */
char *CwNumericLookup[] = {
   "-----",".----","..---","...--","....-",".....","-....",
   "--...","---..","----."
};

struct CwMiscLookupTAG {
   char  Char;
   char  *Morse;
} CwMiscLookup[] = {
   {'?', "..--.."},
   {'/', "-..-."},
   {'.', ".-.-.-"},
   {',', "--..--"},
   {'-', "-...-"},   // aka BT
   {0}   // end of table
};

struct EventHandler {
   char *EventDesc;
   void (*EventHandler)(char *Args);
} EventHandlers[] = {
   { "connected", AnnounceConnect},
   { "disconnected", AnnounceDisconnect},
   { NULL } // end of table
};

void TTSCB(int Pid, int Status)
{
// The TTS program has exited...
   bTTS_Running = FALSE;
}

int StringIsInList(char *List,char *String)
{
   int Ret = FALSE;
   char *cp = strstr(List,String);
   char LastChar;

   if(cp != NULL) do {
      if(cp == List || cp[-1] == ' ' || cp[-1] == ',') {
      // substring starts at the right place
         LastChar = cp[strlen(String)];
         if(LastChar == 0 || LastChar == ' ' || LastChar == ',') {
         // substring ends at the right place
            Ret = TRUE;
         }
      }
   } while(FALSE);
   return Ret;
}

void ExecuteCommand(char *Command)
{
   ConfClient cc;

   memset(&cc,sizeof(cc),0);
   cc.bCmdLine = TRUE;
   cc.bSysop = TRUE;
   cc.bAdmin = TRUE;
   cc.bRfUser = TRUE;
   cc.Callsign = "RF";
   cc.CallPlus = cc.Callsign;
   cc.pCS = piLinkConf;

   LOG_NORM(("RF user executing command \"%s\"\n",Command));
   ConferenceCmd(cc.pCS->pAudio,&cc,Command);
}

void ExpandCommand(char *Command,char *DtmfCommand,char *DtmfWildCards)
{
   char *cp;
   char *cp1;
   int ExtraSpace = 0;
   int WildCardChars = strlen(DtmfWildCards);
   int DtmfChars = strlen(DtmfCommand);
   int bExpandCommand = FALSE;

// Count the space that will be needed when the command is expanded
   cp = Command;
   while((cp = strchr(cp,'$')) != NULL) {
      cp++;
      switch(*cp) {
         case '1':
         // We'll replace "$1" with DtmfCommand
            ExtraSpace -= 2;
            ExtraSpace += DtmfChars;
            bExpandCommand = TRUE;
            break;

         case '2':
         // We'll replace "$2" with DtmfWildCards
            ExtraSpace -= 2;
            ExtraSpace += WildCardChars;
            bExpandCommand = TRUE;
            break;
#if 0
         case '3':
         // We'll replace "$3" with the current port number
            ExtraSpace -= 2;
            ExtraSpace += PortNum > 9 ? 3 : 2;
            break;

         case '$':
         // 
            break;
#endif

         default:
            break;
      }
   }

   if(bExpandCommand != 0) {
   // Something needs expanding
      char *Temp = (char *) malloc(strlen(Command) + ExtraSpace + 1);
      if(Temp != NULL) {
         cp = Command;
         cp1 = Temp;
         while(*cp) {
            if(*cp == '$' && cp[1] == '1') {
            // replace "$1" with DtmfCommand
               strcpy(cp1,DtmfCommand);
               cp1 += DtmfChars;
               cp += 2;
            }
            else if(*cp == '$' && cp[1] == '2') {
            // Replace "$2" with DtmfWildCards
               strcpy(cp1,DtmfWildCards);
               cp1 += WildCardChars;
               cp += 2;
            }
            else {
            // Just copy the character
               *cp1++ = *cp++;
            }
         }

         *cp1 = 0;
      // Execute the expanded command
         ExecuteCommand(Temp);
         free(Temp);
      }
      else {
         LOG_ERROR(("%s#%d: malloc failed.\n",__FUNCTION__,__LINE__));
      }
   }
   else {
   // Nothing to expand, just execute the command as is
      ExecuteCommand(Command);
   }
}

char *CwLookUp(char CwChar)
{
   int   i;
   char *Ret = "";

   if(isalpha(CwChar)) {
      Ret = CwAlphaLookup[toupper(CwChar)-'A'];
   }
   else if(isdigit(CwChar)) {
      Ret = CwNumericLookup[CwChar-'0'];
   }
   else for(i = 0; CwMiscLookup[i].Char != 0; i++) {
      if(CwChar == CwMiscLookup[i].Char) {
         Ret = CwMiscLookup[i].Morse;
         break;
      }
   }

   return Ret;
}

int NodeCompare(const void *avl_a,const void *avl_b,void *avl_param)
{
   char *NodeNameA = ((NamedNode *) avl_a)->NodeName;
   char *NodeNameB = ((NamedNode *) avl_b)->NodeName;
   
   return STRCMP(NodeNameA,NodeNameB);
}

Node *FindNode(char *NodeName)
{
   NamedNode NodeLookup;

   NodeLookup.NodeName = NodeName;
   return (Node *) avl_find(NodeTree,&NodeLookup);
}

// .port <port name>
void CmdSetPort(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   struct avl_traverser avl_trans;
   Node *pNode;
   CPort *pPort;
   NamedNode NodeLookup;

   if(*Arg == 0) {
   // No argument, list ports
      BufPrintf(p,"Available ports:\r");
      pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
      while(pNode != NULL) {
         if(pNode->Enabled()) {
            pPort = dynamic_cast<CPort *> (pNode);
            if(pPort != NULL) {
               pPort->DumpStatus(p);
            }
            else {
               BufPrintf(p,"%c %s\n",CurrentNode == pNode ? '>' : ' ',
                         pNode->NodeName);
            }
         }
         pNode = (Node *) avl_t_next(&avl_trans);
      }
   }
   else {
      NodeLookup.NodeName = Arg;
      if((pNode = (Node *) avl_find(NodeTree,&NodeLookup)) != NULL) {
         CurrentNode = pNode;
      }
      else {
         BufPrintf(p,"Port \"%s\" not found\r",Arg);
         Rcode = TBD_PORT_NOT_FOUND;
      }
   }
}

char *FindToneSpec(int ID)
{
   char *Ret = NULL;
   SavedToneSpec *pSpec = ToneSpecs;

   while(pSpec != NULL) {
      if(pSpec->ID == ID) {
      // Match
         Ret = pSpec->ToneSpec;
         break;
      }
      pSpec = pSpec->Link;
   }
   return Ret;
}

void DtmfCallback(void *user_data, int code, int level)
{
   Node *pClass = (Node *) user_data;

   pClass->SaveDTMF(code == 0 ? ' ' : code);
}

void EndPointEvent(char *EventText)
{
   int Len;
   char *cp;
   int i;

   for(i = 0; EventHandlers[i].EventDesc != NULL; i++) {
      Len = strlen(EventHandlers[i].EventDesc);
      if(strncmp(EventText,EventHandlers[i].EventDesc,Len) == 0) {
         if((cp = strchr(EventText,' ')) != NULL) {
            cp++;
         }

         EventHandlers[i].EventHandler(cp);
         break;
      }
   }
}

// A callsign is entered as using two digits for each letter and number 
// in the callsign. The first digit is the key on which the letter appears 
// (using 1 for Q and Z), and the second digit is 1, 2, or 3, to indicate 
// which letter is being entered.  Numbers are enter as the digit followed 
// by a 0.  The pound key is used to indicate the end of the string.
// For example, the letter "K" is entered as "52", the letter "Q" is 
// entered as "11", and the digit "7" is entered as "70".
//
// Callsigns need not be entered in full.  If a partial callsign is entered, 
// thebridge will find the first match among the stations currently logged on.
// If no match is found among the stations currently logged on, 
// thebridge will say "NOT FOUND".

char DTMF2CharLookup[] = {
   0x21, // 'A'
   0x22, // 'B'
   0x23, // 'C'
   
   0x31, // 'D'
   0x32, // 'E'
   0x33, // 'F'
   
   0x41, // 'G'
   0x42, // 'H'
   0x43, // 'I'

   0x51, // 'J'
   0x52, // 'K'
   0x53, // 'L'

   0x61, // 'M'
   0x62, // 'N'
   0x63, // 'O'

   0x71, // 'P'
   0x11, // 'Q'
   0x72, // 'R'
   0x73, // 'S'

   0x81, // 'T'
   0x82, // 'U'
   0x83, // 'V'

   0x91, // 'W'
   0x92, // 'X'
   0x93, // 'Y'
   0x12, // 'Z'
};

char *DTMF2Call(char *Digit)
{
   static char Callsign[MAX_CALL_LEN+1];
   int   i;
   unsigned int   j;
   char  FirstDigit;
   char  c = 0;
   char  lookup;

   Callsign[0] = 0;

   for(i = 0; i < MAX_CALL_LEN; ) {
   // Get first digit of pair
      c = *Digit++;
      if(!isdigit(c)) {
      // we're done
         break;
      }
      FirstDigit = c;

   // Get second digit of pair
      c = *Digit++;
      if(!isdigit(c)) {
      // we're done
         break;
      }
      
      if(c == '0') {
      // digit
         Callsign[i] = FirstDigit;
      }
      else {
         lookup = ((FirstDigit << 4) & 0xf0) | (c & 0xf);
         for(j = 0; j < sizeof(DTMF2CharLookup); j++) {
            if(DTMF2CharLookup[j] == lookup) {
               Callsign[i] = (char) j + 'A';
               break;
            }
         }

         if(j == sizeof(DTMF2CharLookup)) {
         // invalid combination
            Callsign[0] = 0;
            break;
         }
      }
   }

   if(c != 0 && c != '#') {
   // invalid combination
      Callsign[0] = 0;
   }
   else {
   // Add terminating null
      Callsign[i] = 0;
   }

   return Callsign;
}

void CmdSayStatus(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   ConfClient *pCC;
   int Connections = piLinkConf->ConfTree->avl_count;
   int Announced = 0;
   struct avl_traverser avl_trans;
   
   if(Connections > 0) {
      CurrentNode->SayPhrase("I am currently connected to ");
      if(Connections > 1) {
         CurrentNode->SayPhrase("the following stations ");
      }
      pCC = (ConfClient *) avl_t_first(&avl_trans,piLinkConf->ConfTree);
      while(pCC != NULL) {
         if(Connections > 1 && (Connections - Announced) == 1) {
            CurrentNode->SayPhrase("and ");
         }
         CurrentNode->SayCall(pCC->Callsign);
         pCC = (ConfClient *) avl_t_next(&avl_trans);
      }
   }

   CurrentNode->SayPhrase("I have been running for ");
   CurrentNode->SayPhrase(TimeLapse2String(BootTime,FALSE));
}

void ScriptExitCB(int Pid, int Status)
{  
   // Nothing to do, this is just here to prevent creating a zombie army
}

void CmdScript(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
#define MAX_ARGS  16
   char *argv[MAX_ARGS];
   pid_t Pid = -1;
   char  *cp = Arg;
   int   i = 0;

// NB: make sure all buffered streams are written to disk,
// otherwise the data will be written twice, once by the main
// process and once by the forked process !

   if(fflush(NULL)) {
      LOG_ERROR(("%s: fflush failed %s",__FUNCTION__,Err2String(errno)));
   }

   while(i < MAX_ARGS && cp != 0) {
      argv[i++] = cp;
      if((cp = strchr(cp,' ')) != NULL) {
         *cp++ = 0;
         while(isspace(*cp)) {
            cp++;
         }
      }
   }

   if(i == MAX_ARGS) {
      BufPrintf(p,"too many arguments\r");
      LOG_ERROR(("%s: error: too many arguments\n",__FUNCTION__));
   }
   else {
      argv[i] = NULL;
      if((Pid = fork()) < 0) {
         LOG_ERROR(("%s: fork failed %s",__FUNCTION__,Err2String(errno)));
      }
      else if(Pid == 0) {
         int Error;
      // child
         setenv("TLB_PORT",CurrentNode->NodeName,TRUE);
         Error = execvp(argv[0],argv);
         exit(1);
      }
      else {
      // Parent
         LOG_ERROR(("%s: script \"%s\"started, Pid %d\n",__FUNCTION__,Arg,
                    (int) Pid));
         RegisterExitCB(Pid,ScriptExitCB);
      }
   }
#undef MAX_ARGS
}

// <To> <From> [<From>] ...
// <To> and <From> can be a Voip connection (callsign) or a port name
void DoLinks(ClientInfo *p,char *Arg,int bLink,int bMonitor)
{
#define MAX_ARGS  16
   char *PortNames[MAX_ARGS];
   Node *TargetNode[MAX_ARGS];
   char  *cp = Arg;
   int i;
   int ArgCount = 0;
   Node *pToNode;

   memset(TargetNode,0,sizeof(TargetNode));

   do {
      while(ArgCount < MAX_ARGS && *cp) {
         PortNames[ArgCount++] = cp;
         while(*cp && !(isspace(*cp) || *cp == ',')) {
            cp++;
         }

         while(*cp && (isspace(*cp) || *cp == ',')) {
            *cp++ = 0;
         }
      }

      if(ArgCount == MAX_ARGS) {
         BufPrintf(p,"Error: too many arguments\r",cp);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      for(i = 0; i < ArgCount; i++) {
      // Find specified port or Voip connection
         if((TargetNode[i] = FindNode(PortNames[i])) == NULL) {
         // node not found
            BufPrintf(p,"Error: can't find port or connection named \"%s\"\r",
                      PortNames[i]);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }
      }

      if(Rcode != 0) {
         break;
      }

      pToNode = TargetNode[0];

      if(pToNode == NULL) {
         LOG_ERROR(("%s#%d: Internal error!\n",__FUNCTION__,__LINE__));
         BufPrintf(p,"Internal error @ %s#%d\r",__FUNCTION__,__LINE__);
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      if(ArgCount == 1 && !bLink) {
      // unlink everything from specified port/connection
         if(pToNode->UnlinkAll()) {
            BufPrintf(p,"Internal error @ %s#%d\r",__FUNCTION__,__LINE__);
            Rcode = TBD_ERR_INVALID_ARG;
         }
         break;
      }

      if(ArgCount < 2) {
         BufPrintf(p,"Error: not enough arguments\r");
         Rcode = TBD_ERR_INVALID_ARG;
         break;
      }

      BufPrintf(p,"%slinking ",bLink ? "" : "un");

      for(i = 1; i < ArgCount; i++) {
         if(bLink) {
            BufPrintf(p,"%s%s",i > 1 ? ", " : "",TargetNode[i]->NodeName);
            pToNode->AddRx(TargetNode[i]);
            if(!bMonitor) {
               TargetNode[i]->AddRx(pToNode);
            }
         }
         else {
            pToNode->DeleteRx(TargetNode[i]);
         }
      }
      if(bLink) {
         BufPrintf(p," to %s%s\r",PortNames[0],bMonitor ? " (monitor only)":"");
      }
      else {
         BufPrintf(p," from %s\r",PortNames[0]);
      }
   } while(FALSE);
#undef MAX_ARGS
}

void CmdLink(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   int bFirst;

   if(*Arg) {
      if(Arg[0] == '-') {
         switch(Arg[1]) {
            case 'm':
            // Monitor command
               Arg += 2;
               while(*Arg == ' ') {
                  Arg++;
               }
               DoLinks(p,Arg,TRUE,TRUE);
               break;

            case 'h':
            case '?':
               BufPrintf(p,"Usage: .link [-m] <destination> <source> "
                            "[<source>...]\r");
               Rcode = TBD_ERR_INVALID_ARG;
               break;

            default:
               BufPrintf(p,"Error: Unknown switch \"%c\"\r",Arg[1]);
               Rcode = TBD_ERR_INVALID_ARG;
               break;
         }
      }
      else {
         DoLinks(p,Arg,TRUE,FALSE);
      }
   }
   else {
   // No argument, list linked ports
      BufPrintf(p,"Port mapping:\n");
      struct avl_traverser avl_trans;
      LinkedNode *pLink;

      Node *pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
      while(pNode != NULL) {
         if(pNode->Enabled()) {
            BufPrintf(p,"%10s: ",pNode->NodeName);
            bFirst = TRUE;

            pLink = pNode->pReceivers;
            while(pLink != NULL) {
               if(pLink->pRx->Enabled()) {
                  BufPrintf(p,"%s%s",bFirst ? "" : ", ",pLink->pRx->NodeName);
                  bFirst = FALSE;
               }
               pLink = pLink->pNextRx;
            }
            BufPrintf(p,"%s\r",bFirst ? "none" : "");
         }
         pNode = (Node *) avl_t_next(&avl_trans);
      }
   }
}

void CmdUnLink(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   Node *pNode;
   struct avl_traverser avl_trans;
   int UnLinkAll = 0;
   const char *UnlinkMsg[] = {
      "ports and voip connections",
      "ports",
      "voip connections",
   };

   if(*Arg == 0) {
   // No argument
      BufPrintf(p,"usage:\r");
      BufPrintf(p,"unlink all\r");
      BufPrintf(p,"unlink rf\r");
      BufPrintf(p,"unlink voip\r");
      BufPrintf(p,"unlink <destination> [<source>...]\r");
      Rcode = TBD_ERR_INVALID_ARG;
   }
   else if(STRCMP(Arg,"all") == 0) {
      UnLinkAll = 1;
   }
   else if(STRCMP(Arg,"rf") == 0) {
      UnLinkAll = 2;
   }
   else if(STRCMP(Arg,"voip") == 0) {
      UnLinkAll = 3;
   }

   if(UnLinkAll != 0) {
      pNode = (Node *) avl_t_first(&avl_trans,NodeTree);
      while(pNode != NULL) {
         if(UnLinkAll == 1 ||
            (UnLinkAll == 2 && dynamic_cast<CPort *> (pNode) != NULL) || 
            (UnLinkAll == 3 && dynamic_cast<VoipNode *> (pNode) != NULL))
         {
            if(pNode->UnlinkAll()) {
               BufPrintf(p,"Internal error @ %s#%d\r",__FUNCTION__,__LINE__);
               Rcode = TBD_ERR_INVALID_ARG;
               break;
            }
         }
         pNode = (Node *) avl_t_next(&avl_trans);
      }
      BufPrintf(p,"all %s unlinked\r",UnlinkMsg[UnLinkAll-1]);
   }
   else {
      DoLinks(p,Arg,FALSE,FALSE);
   }
}

NamedNode::~NamedNode() 
{
}

Node::Node()
{
   AudioAvailable = 0;
   bCwToneOn = FALSE;
   bElementSpace = FALSE;
   bRxActive = 0;
   bRxTimeout = FALSE;
   bTxTimeOut = 0;
   CommandFilename = NULL;
   CTCSSAngle = 0.0;
   CTCSSLevel = 0;
   Ctone = 0;
   CtoneDelay = 0;
   CtoneMap = NULL;
   CtoneResetsTimeout = 0;
   CtoneTimer.tv_sec = 0;
   CWAutoJam = 0;
   CWChar = NULL;
   CWSpeed = 20;
   CWToneLevel = 5000;
   Dtmf1stDigits = NULL;
   DTMF_Digits = 0;
   DtmfCoverTone = 0;
   DtmfMethod = 0;
   EchoCTCSSFreq = 0.0;
   Element = NULL;
   ElementCount = 0;
   LastDTMFDigit.tv_sec = 0;
   LastMixGain = 1024;  // 1:1
   Min1stDigitHoldTime = 0;
   MinDtmfInterdigitTime = 0;
   MuteDTMF = 1;
   NodeName = NULL;
   pDTMFCommands = NULL;
   Playfp = NULL;
   pPlayHead = NULL;
   pReceivers = NULL;
   pTransmitters = NULL;
   RampCount = 0;
   ReverseTwist = -1;
   RxTimeout = 0;
   RxTimeoutTone = 0;
   ToneGenerators = NULL;
   Transmitters = 0;
   Twist = -1;
   TxKeyTime = 0;
   TxTimeout = 0;
   TxTimeoutTone = 0;
   TxTimeoutToneGen = NULL;
   BufNum = 1;
   bCwIsID = 0;
   bAnnoucementIsID = 0;
   PrefixedTone = 0;
   DTMFPrefix = NULL;
   DTMFCommandTimeout = 1000;    // 1 second
   PrefixTimeout = 8000;         // 8 seconds
   bPrefixed = 0;
}

Node::~Node()
{
   UnlinkAll();
   if(NodeName != NULL) {
      free(NodeName);
   }

   if(CurrentNode == this) {
      CurrentNode = NULL;
   }
}

int Node::UnlinkAll()
{
   int Ret = 0;
   LinkedNode *pLink;

// unlink everything from this node
   while((pLink = pReceivers) != NULL) {
      pReceivers = pLink->pNextRx;
      if(pLink->pRx->DeleteTx(pLink)) {
      // Shouldn't happen !
         LOG_ERROR(("%s#%d: Internal error!\n",__FUNCTION__,__LINE__));
         Ret = 1;
      }
      delete pLink;
   }

   while((pLink = pTransmitters) != NULL) {
      pTransmitters = pLink->pNextTx;
      if(pLink->pTx->DeleteRx(pLink->pRx)) {
      // Shouldn't happen !
         LOG_ERROR(("%s#%d: Internal error!\n",__FUNCTION__,__LINE__));
         Ret = 1;
      }
      delete pLink;
   }

   return Ret;
}

void Node::FreeAnnoucement()
{
   PlayPCM *p;

   if(Playfp != NULL) {
      fclose(Playfp);
      Playfp = NULL;
   }

   if((p = pPlayHead) != NULL) {
      pPlayHead = p->Link;
      if(p->ReplyPort != 0) {
         ConfClient *pCC = &CmdLineCC;
         ClientInfo *pCInfo = CmdClient;
      // Restore the saved reply port.  
      // (we can't just use pCC->HisAdr.PORT since another client may have 
      // sent a command since the client that started the playback.)
         pCC->HisAdr.PORT = p->ReplyPort;
         pCInfo->Count = 0;
         BufPrintf(pCInfo,"%d\nPlayback of %s complete\n",TBD_SAY_COMPLETE,
                   p->FileName);
         sendto(pCInfo->Socket,pCInfo->Buf,pCInfo->Count+1,0,&pCC->HisAdr.s,
                sizeof(IPAdrUnion));
      }
      free(p->FileName);
      free(p);
   }
}

void Node::StartNextAnnoucement()
{
   PlayPCM *p = pPlayHead;
   FILE *TTSInFp = NULL;
   char Filename[PATH_MAX];
   int Bytes2Write;
   char *argv[4];
   char OutFilename[PATH_MAX];
   pid_t TTSPid = -1;
   int ReplyPort = 0;
   
   while((p = pPlayHead) != NULL) {
      if(p->Flags & PLAY_FLAG_TTS) {
         if(TTSInFp == NULL) {
         // Open TTS input file
            snprintf(Filename,PATH_MAX,"%s/%s",WorkingDir,TTS_IN_FILENAME);
            if((TTSInFp = fopen(Filename,"w")) == NULL) {
               LOG_ERROR(("%s#%d: Couldn't open \"%s\", %s",__FUNCTION__,
                          __LINE__,Filename,Err2String(errno)));
            }
         }
         else {
         // File already open, append a space so we don't run words together
            if(fputs(" ",TTSInFp) == EOF) {
               LOG_ERROR(("%s: Error reading \"%s\", %s",__FUNCTION__,
                          Filename,Err2String(errno)));
               fclose(TTSInFp);
               TTSInFp = NULL;
            }
         }

         if(TTSInFp != NULL) {
            Bytes2Write = strlen(p->FileName);
            if(fwrite(p->FileName,Bytes2Write,1,TTSInFp) != 1) {
               LOG_ERROR(("%s: Error reading \"%s\", %s",__FUNCTION__,
                          Filename,Err2String(errno)));
               fclose(TTSInFp);
               TTSInFp = NULL;
            }
         }
         if(ReplyPort == 0) {
         // Forward ReplyPort to the playback of the generated wave file
            ReplyPort = p->ReplyPort;
         }
         p->ReplyPort = 0;
         FreeAnnoucement();
      }
      else {
         if(TTSInFp != NULL) {
         // Don't start PCM file playback until TTS finishes
            break;
         }
         if((Playfp = fopen(p->FileName,"r")) == NULL) {
            LOG_ERROR(("%s#%d: Couldn't open \"%s\", %s",__FUNCTION__,__LINE__,
                       p->FileName,Err2String(errno)));
            FreeAnnoucement();
         }
         else {
            WaveHdr Hdr;
         // read 44 byte WAV header
            if(fread(&Hdr,sizeof(Hdr),1,Playfp) != 1) {
               LOG_ERROR(("%s: fread failed reading %s - %s\n",__FUNCTION__,
                          p->FileName,Err2String(errno)));
               FreeAnnoucement();
            }
            else {
            // Stupid test for now... fix me someday
               if(Hdr.wBitsPerSample == 8) {
                  b8BitFile = TRUE;
                  break;
               }
               else if(Hdr.wBitsPerSample == 16) {
                  b8BitFile = FALSE;
                  break;
               }
               else {
                  LOG_ERROR(("%s: %s has unsupported wBitsPerSample (%d)\n",
                             __FUNCTION__,p->FileName,Hdr.wBitsPerSample));
                  FreeAnnoucement();
               }
            }
         }
      }
   }

   if(TTSInFp != NULL) {
   // An TTS input file has been written, start the TTS program to 
   // render it to audio
      fclose(TTSInFp);

      snprintf(OutFilename,PATH_MAX,"%s/%s",WorkingDir,TTS_OUT_FILENAME);
      unlink(OutFilename);
   // Create a new announcement entry at the front of the queue for playback

      PlayAnnouncementFile(OutFilename,PLAY_FLAG_TTS_OUT,ReplyPort);

   // NB: make sure all buffered streams are written to disk,
   // otherwise the data will be written twice, once by the main
   // process and once by the forked process !

      if(fflush(NULL)) {
         LOG_ERROR(("%s: fflush failed %s",__FUNCTION__,Err2String(errno)));
      }

      bTTS_Running = TRUE;
      if((TTSPid = fork()) < 0) {
         FreeAnnoucement();
         bTTS_Running = FALSE;
         LOG_ERROR(("%s: fork failed %s",__FUNCTION__,Err2String(errno)));
      }
      else if(TTSPid == 0) {
      // child
         argv[0] = TTS_ExePath;
         argv[1] = Filename;
         argv[2] = OutFilename;
         argv[3] = NULL;
         execvp(TTS_ExePath,argv);
      // NB: if execvp fails, BAIL!
         exit(1);
      }
      else {
      // Parent
         RegisterExitCB(TTSPid,TTSCB);
      }
   }

   if(pPlayHead != NULL && pPlayHead->Flags & PLAY_FLAG_ID) {
   // This annoucement is an voice ID 
      bAnnoucementIsID = TRUE;
   }
   else {
      bAnnoucementIsID = FALSE;
   }
}


void Node::PlayAnnouncementFile(char *FileName,int Flags,int ReplyPort)
{
   PlayPCM *p;

   if((p = (PlayPCM *) malloc(sizeof(PlayPCM))) == NULL) {
      LOG_ERROR(("PlayAnnouncementFile(): Error: out of memory\n"));
   }
   else {
      p->FileName = strdup(FileName);
      p->Flags = Flags;
      p->ReplyPort = ReplyPort;
      if(Flags & PLAY_FLAG_TTS_OUT) {
      // put on head of queue
         p->Link = pPlayHead;
         pPlayHead = p;
         if(p->Link == NULL) {
            pPlayTail = p;
         }
      }
      else {
      // put on tail of queue
         p->Link = NULL;
         if(pPlayHead == NULL) {
            pPlayHead = p;
         }
         else {
            pPlayTail->Link = p;
         }
         pPlayTail = p;
      }
      if(Playfp == NULL && !(Flags & PLAY_FLAG_TTS_OUT)) {
         StartNextAnnoucement();
      }
   }
}

// Say the specified phrase
void Node::SayPhrase(char *Phrase,int ReplyPort)
{
   PlayPCM *p;
   char FileName[PATH_MAX];
   char *cp;
   char *Temp;
   struct stat FileStats;

   FileName[0] = 0;

   if((cp = strstr(Phrase,".wav")) != NULL && strlen(cp) == 4) {
   // a filename was specified
      strcpy(FileName,Phrase);
   }
   else if(TTS_ExePath == NULL) {
   // Lookup PCM file that contains the Phrase we want to say
      if(strlen(Phrase) == 1) {
      // Special case just a single character
         snprintf(FileName,sizeof(FileName),"pcm/%c.wav",toupper(*Phrase));
      }
      else {
      // replace all spaces with underscores
         Temp = strdup(Phrase);
         cp = Temp;
         while((cp = strchr(cp,' ')) != NULL) {
            *cp++ = '_';
         }
         snprintf(FileName,sizeof(FileName),"pcm/%s.wav",Temp);
         free(Temp);
         if(stat(FileName,&FileStats) != 0) {
         // no such file
            FileName[0] = 0;
            LOG_ERROR(("%s: Unable to find file for phrase \"%s\"\n",
                       __FUNCTION__,Phrase));
         }
      }
   }

   if(TTS_ExePath != NULL || FileName[0] != 0) {
      if((p = (PlayPCM *) malloc(sizeof(PlayPCM))) == NULL) {
         LOG_ERROR(("%s: Error: out of memory\n",__FUNCTION__));
      }
      else {
         p->Link = NULL;
         p->ReplyPort = ReplyPort;
         if(pPlayHead == NULL) {
            pPlayHead = p;
         }
         else {
            pPlayTail->Link = p;
         }
         pPlayTail = p;
         if(FileName[0] != 0) {
            p->Flags = 0;
            p->FileName = strdup(FileName);
            if(Playfp == NULL) {
               StartNextAnnoucement();
            }
         }
         else {
            p->Flags = PLAY_FLAG_TTS;
            p->FileName = strdup(Phrase);
         }
      }
   }

   CPort *pPort = dynamic_cast<CPort *> (this);

   if(pPort != NULL && !pPort->pAudioC->bWriteWait && Playfp != NULL) {
      WriteAudioOut(pPort->pAudioC);
   }
}

void Node::SayCall(char *Call)
{
   int i;
   char FileName[80];
   int CallLen = strlen(Call);
   char Temp[80];
   char  *cp;
   struct stat FileStats;
   char AlphaNumeric[2];

// Try for a exact match filename

   snprintf(Temp,sizeof(Temp),"%s",*Call == '*' ? &Call[1] : Call);
   if((cp = strchr(Temp,'*')) != NULL) {
   // remove trailing '*'
      *cp = 0;
   }
   if((cp = strchr(Temp,' ')) != NULL) {
   // remove trailing spaces
      *cp = 0;
   }
   snprintf(FileName,sizeof(FileName),"pcm/%s.wav",Temp);
   if(stat(FileName,&FileStats) == 0) {
   // we have a file for this callsign/conference name, play it.
      PlayAnnouncementFile(FileName,0);
   }
   else {
   // Don't have an exact match, spell the call letter by letter
      for(i = 0; i < CallLen; i++) {
         if(isalnum(Call[i])) {
            AlphaNumeric[0] = Call[i];
            AlphaNumeric[1] = 0;
            SayPhrase(AlphaNumeric);
         }
         else if(Call[i] == '*') {
            if(i != 0) {
               SayPhrase("conference");
            }
         }
         else if(Call[i] == '-') {
            i++;
            if(toupper(Call[i]) == 'L') {
            // Link
               SayPhrase("link");
            }
            else if(toupper(Call[i]) == 'R') {
            // Repeater
               SayPhrase("repeater");
            }
         }
         else if(Call[i] == ' ') {
            break;
         }
      }
   }

   CPort *pPort = dynamic_cast<CPort *> (this);

   if(pPort != NULL && !pPort->pAudioC->bWriteWait) {
      pPort->WriteAudioOut(pPort->pAudioC);
   }
}

// "connected <protocol> <callsign> <number of connections>"
void AnnounceConnect(char *Args)
{
   char *cp;

   if((cp = strchr(Args,' ')) != NULL) {
      if(CurrentNode != NULL) {
         CurrentNode->SayPhrase("connected to");
         CurrentNode->SayCall(cp+1);
      }
   }
}

// disconnected <reason> <callsign> <number of connections left>"
void AnnounceDisconnect(char *Args)
{
   char *cp;
   char *cp1;

   if((cp = strchr(Args,' ')) != NULL) {
      cp++;
      if((cp1 = strchr(cp,' ')) != NULL) {
         *cp1 = 0;
      }

      if(StationDisconnected != NULL) {
      // suppress the announcement if we explicitly disconnected this guy
         if(STRCMP(StationDisconnected,cp) == 0) {
         // We disconnected him
            cp = NULL;
            free(StationDisconnected);
            StationDisconnected = NULL;
         }
      }
   }

   if(cp != NULL && CurrentNode != NULL) {
      CurrentNode->SayCall(cp);
      CurrentNode->SayPhrase("disconnected");
   }
}
// Put PCM for audio annoucements into static AudioBuf
int Node::GetAnnounceAudio()
{
   int Ret = FALSE;
   int Read;
   int Bytes2Read;
   
   PlayAvail = 0;

   while(!bTTS_Running && pPlayHead != NULL && PlayAvail < AUDIO_BUF_SIZE) {
      if(Playfp == NULL) {
         StartNextAnnoucement();
      }

      if(Playfp != NULL) {
         if(b8BitFile) {
            unsigned char TempBuf[AUDIO_BUF_SIZE];
            int i;

            Bytes2Read = sizeof(TempBuf) - PlayAvail;
            Read = fread(TempBuf,1,Bytes2Read,Playfp);
            if(Read > 0) {
            // Convert to 16 bit samples in AudioBuf
               for(i = 0; i < Read; i++) {
                  AudioBuf[PlayAvail++] = (TempBuf[i] - 0x80) << 8;
               }
            }
         }
         else {
            Bytes2Read = sizeof(AudioBuf) - (PlayAvail * sizeof(int16));
            Read = fread(&AudioBuf[PlayAvail],1,Bytes2Read,Playfp);
            if(Read > 0) {
               PlayAvail += Read / sizeof(int16);
            }
         }

         if(Read > 0) {
            Ret = TRUE;
            if(PlayAvail == AUDIO_BUF_SIZE) {
               break;
            }
         }
         else {
            if(!feof(Playfp)) {
               LOG_ERROR(("GetAnnounceAudio(): read error : %s",
                          Err2String(errno)));
            }
            FreeAnnoucement();
         }
      }
   }

   if(PlayAvail > 0 && PlayAvail < AUDIO_BUF_SIZE) {
   // Last announcement in the chain did not fill buffer, zero the remainder
      memset(&AudioBuf[PlayAvail],0,(AUDIO_BUF_SIZE-PlayAvail) * sizeof(int16));
      PlayAvail = AUDIO_BUF_SIZE;
   }

   return Ret;
}

void Node::MixAudio(int16 *In,int Samples)
{
   int *Out = MixBuffer;
   int i;

   for(i = 0; i < Samples; i++) {
      *Out++ += *In++;
   }
}

void Node::UpdateCtoneFlags()
{
   if(CtoneMap != NULL) {
      LinkedNode *pLink = pTransmitters;
      while(pLink != NULL) {
         if(StringIsInList(CtoneMap,pLink->pTx->NodeName)) {
            pLink->Flags |= LNK_FLG_CTONE;
         }
         else {
            pLink->Flags &= ~LNK_FLG_CTONE;
         }
         pLink = pLink->pNextTx;
      }
   }
}

void Node::AddTx(LinkedNode *pNewLink)
{
   LinkedNode *pLink = pTransmitters;

// Make sure this transmitter isn't already linked to this receiver
   while(pLink != NULL) {
      if(pLink->pTx == pNewLink->pTx) {
      // Already linked
         LOG_ERROR(("%s: Error: %s already linked to %s\n",__FUNCTION__,
                    pLink->pRx->NodeName,pLink->pTx->NodeName));
         break;
      }
      pLink = pLink->pNextTx;
   }

   if(pLink == NULL) {
      pNewLink->pNextTx = pTransmitters;
      pTransmitters = pNewLink;
      Transmitters++;
      UpdateCtoneFlags();
   }
}

int Node::DeleteTx(LinkedNode *pDeleteLink)
{
   LinkedNode *pLastLink = NULL;
   LinkedNode *pLink = pTransmitters;

   while(pLink != NULL) {
      if(pLink == pDeleteLink) {
         if(pLastLink == NULL) {
            pTransmitters = pLink->pNextTx;
         }
         else {
            pLastLink->pNextTx = pLink->pNextTx;
         }
         Transmitters--;
         break;
      }
      pLastLink = pLink;
      pLink = pLink->pNextTx;
   }

   return (pLink == NULL);
}

void Node::AddRx(Node *pRx)
{
   LinkedNode *pLink = pReceivers;

// Make sure this receiver isn't already linked to this transmitter
   while(pLink != NULL) {
      if(pLink->pRx == pRx) {
      // Already linked
         break;
      }
      pLink = pLink->pNextRx;
   }

   if(pLink == NULL) {
      if((pLink = new LinkedNode) == NULL) {
         LOG_ERROR(("%s#%d: new failed!\n",__FUNCTION__,__LINE__));
      }
      else {
         memset(pLink,0,sizeof(LinkedNode));
         pLink->pRx = pRx;
         pLink->pTx = this;
         pRx->AddTx(pLink);
      // TODO: insert into link in sorted order so links display nicely
         pLink->pNextRx = pReceivers;
         pReceivers = pLink;
      }
   }
}

int Node::DeleteRx(Node *pRx)
{
   LinkedNode *pLink;
   LinkedNode *pLastLink = NULL;
   
   pLink = pReceivers;
   while(pLink != NULL) {
      if(pLink->pRx == pRx) {
         if(pLastLink == NULL) {
            pReceivers = pLink->pNextRx;
         }
         else {
            pLastLink->pNextRx = pLink->pNextRx;
         }
         break;
      }
      pLastLink = pLink;
      pLink = pLink->pNextRx;
   }

   return (pLink == NULL);
}

// Scale the mixed audio so that the maximum level does not exceed 
// MAX_MIXED_LEVEL.  When the mix gain changes try to change the gain
// at a zero crossing to avoid thumps.
void Node::ScaleMixedAudio()
{
   int MixGain;
   int i;
   int bFindNegCrossing = MixBuffer[0] > 0 ? TRUE : FALSE;
   int Sample;

   MaxMixedLevel = 0;
   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      if((Sample = MixBuffer[i]) < 0) { 
         Sample = -Sample;
      }
      if(Sample > MaxMixedLevel) {
         MaxMixedLevel = Sample;
      }
   }

   if(MaxMixedLevel > MAX_MIXED_LEVEL) {
   // calculate gain to reduce maximum mixed level to 
      MixGain = (MAX_MIXED_LEVEL * 1024) / MaxMixedLevel;
   }
   else {
      MixGain = 1024;   // 1:1
   }

   if(MixGain == 1024 && LastMixGain == 1024) {
   // Both this gain and the last gain were 1:1, so no need to multiply
      for(i = 0; i < AUDIO_BUF_SIZE; i++) {
         AudioBuf[i] = (int16) MixBuffer[i];
      }
   }
   else {
      if((MixGain - LastMixGain) > 2) {
      // increase gain slowly to make it less noticeable
         MixGain = LastMixGain + (MixGain - LastMixGain)/2;
      }

      for(i = 0; i < AUDIO_BUF_SIZE; i++) {
         if(MixBuffer[i] > MAX_MIXED_LEVEL || MixBuffer[i] < -MAX_MIXED_LEVEL) {
         // need to apply new MixGain now
            break;
         }

         if(bFindNegCrossing) {
            if(MixBuffer[i] < 0) {
            // apply new gain now
               break;
            }
         }
         else {
            if(MixBuffer[i] > 0) {
            // apply new gain now
               break;
            }
         }

      // AudioBuf[i] = MixBuffer[i] * LastMixGain / 1024;
         AudioBuf[i] = (int16) ((MixBuffer[i] * LastMixGain) >> 10);
      }

   // change to the new mix gain for the remainder of the buffer
      LastMixGain = MixGain;

      for(; i < AUDIO_BUF_SIZE; i++) {
         AudioBuf[i] = (int16) ((MixBuffer[i] * LastMixGain) >> 10);
      }
   }
}

void Node::GetTxAudio()
{
   CToneGen *pToneGen;
   CToneGen *pLastGen;
   CToneGen *pNextGen;
   LinkedNode *pLink;
   int Sources = 0;
   int IDSources = 0;
   int Samples;

   AudioAvailable = 0;
   memset(MixBuffer,0,sizeof(MixBuffer));

   if(GetAnnounceAudio()) {
      if(bAnnoucementIsID) {
         IDSources++;
      }
      Sources++;
      MixAudio(AudioBuf,AUDIO_BUF_SIZE);
   }

   if(CWChar != NULL) {
      if(bCwIsID) {
         IDSources++;
      }
      Sources++;
      GenCW(AudioBuf);
      MixAudio(AudioBuf,AUDIO_BUF_SIZE);
   }

   pToneGen = ToneGenerators;
   pLastGen = NULL;
   while(pToneGen != NULL) {
      pNextGen = pToneGen->Link;
      if(pToneGen->ToneGen(AudioBuf,AUDIO_BUF_SIZE)) {
      // This generator has finished
         if(pLastGen == NULL) {
            ToneGenerators = pNextGen;
         }
         else {
            pLastGen->Link = pNextGen;
         }
         delete pToneGen;
      }
      else {
         pLastGen = pToneGen;
      }
      Sources++;
      MixAudio(AudioBuf,AUDIO_BUF_SIZE);
      pToneGen = pNextGen;
   }

   pLink = pReceivers;
   while(pLink != NULL) {
      int16 *pAudio;

      if((Samples = pLink->pRx->GetRxAudio(pLink,&pAudio)) > 0) {
         Sources++;
         MixAudio(pAudio,Samples);
      }
      pLink = pLink->pNextRx;
   }

   if(Sources > 0 && CTCSSLevel > 0) {
      Sources++;
      IDSources++;
      GenCTCSS(AudioBuf);
      MixAudio(AudioBuf,AUDIO_BUF_SIZE);
   }

// Now apply the mix gain to prevent over modulation

   if(Sources > 0) {
      ScaleMixedAudio();
      AudioAvailable = sizeof(AudioBuf);
      if(Sources > IDSources) {
      // A source other than an ID is active, update the LastActivity time
         LastActivity = TimeNow.tv_sec;
      }
   }

   if(bTxTimeOut) {
   // Only return audio from the Timeout tone if timed out
      if(TxTimeoutToneGen != NULL) {
         AudioAvailable = sizeof(AudioBuf);
         if(TxTimeoutToneGen->ToneGen(AudioBuf,AUDIO_BUF_SIZE)) {
         // This generator has finished
            D2PRINTF(("%s: Deleting timeout tonegen #%d\n",__FUNCTION__,
                      TxTimeoutToneGen->ID));
            delete TxTimeoutToneGen;
            TxTimeoutToneGen = NULL;
         }
      }
      else {
         AudioAvailable = 0;
      }
   }
}

int Node::WriteAudioOut(ClientInfo *p)
{
   return 0;
}

void Node::DecodeDTMF(char *Command)
{
   DTMF_Command *p = pDTMFCommands;
   char  WildCardBuf[MAX_DTMF_DIGITS+1];
   int i;
   int j;
   int WildCardCount = 0;
   char *cp;
   char *CmdPtr;
   int Digits = strlen(Command);

   CurrentNode = this;

   LOG_NORM(("%s: Decoding \"%s\"\n",__FUNCTION__,Command));

   while(p != NULL) {
      for(i = 0, j = 0; i < Digits; i++) {
         if(p->DTMF_String[j] == '?' || p->DTMF_String[j] == '+') {
         // Wildcard match, save the matched character
            WildCardBuf[WildCardCount++] = Command[i];
            if(p->DTMF_String[j] == '?') {
               j++;
            }
         }
         else if(p->DTMF_String[j] == Command[i]) {
         // match
            j++;
         }
         else {
         // no match
            break;
         }
      }

      if(i == Digits && (p->DTMF_String[j] == 0 || p->DTMF_String[j] == '+')) {
      // Found a matching command
         break;
      }
   // No match, keep looking
      WildCardCount = 0;
      p = p->Link;
   }

   if(p != NULL) {
   // Found the command, execute it
      char *Temp;
      WildCardBuf[WildCardCount] = 0;

   // Make a copy of the command in case the command causes a new
   // command file to be loaded (which will free the passed in string).
      Temp = strdup(p->Command);
      CmdPtr = Temp;
      while(CmdPtr != NULL) {
         if((cp = strchr(CmdPtr,';')) != NULL) {
            *cp = 0;
         }
      // Expand and execute the command
         ExpandCommand(CmdPtr,Command,WildCardBuf);

         if(cp != NULL) {
         // restore the command seperator character we removed earlier
            *cp++ = ';';
         // Skip any white space to make the logs look pretty
            while(isspace(*cp)) {
               cp++;
            }
         }

         CmdPtr = cp;
      }
      free(Temp);
   }
   else {
      LOG_ERROR(("Failed to decode touchtone command \"%s\"\n",Command));
   }
}

// Return:
// 1  - prefix match
// 0  - prefix incomplete
// -1 - prefix match fail
int Node::CheckPrefix()
{
   int Ret = 0;
   int i;

   printf("checking prefix against \"%s\"\n",DTMF_Buf);
   for(i = 0; i < DTMF_Digits; i++) {
      if(DTMF_Buf[i] != DTMFPrefix[i]) {
         printf("Prefix digit %d '%c' match failed\n",i,DTMF_Buf[i]);
         Ret = -1;
         break;
      }
      printf("Prefix digit %d '%c' matched\n",i,DTMF_Buf[i]);
   }

   if(DTMFPrefix[i] == 0) {
      Ret = 1;
   }
   printf("%s: returing %d\n",__FUNCTION__,Ret);
   return Ret;
}

void Node::SaveDTMF(char Digit)
{
   int bIgnoreDigit = FALSE;
   int Elapsed = TimeLapse(&LastDTMFDigit);

   if(DtmfCTCSSMode == 0 ||
      DtmfCTCSSMode == 1 && bRxActive || 
      DtmfCTCSSMode == 2 && bCtcssActive)
   {
      if(Digit == ' ') {
      // End of decode
         if(DTMF_Digits == 1 && Min1stDigitHoldTime != 0) {
         // End of decode of first digit
            if(Elapsed < Min1stDigitHoldTime) {
            // First digit wasn't held long enough, ignore it
               LOG_ERROR(("%s: Ignoring 1st digit \'%c\', only held %d ms.\n",
                          __FUNCTION__,DTMF_Buf[0],Elapsed));
               DTMF_Digits = 0;
               if(MuteDTMF && DtmfCoverTone != 0) {
                  CancelToneGen(DtmfCoverTone);
               }
            }
         }
      }
      else {
      // DTMF Digit decoded 
         DPRINTF(("Port %s TouchTone decode: %c\n",NodeName,Digit));
         DFLUSH();
         if(DTMF_Digits > 0 && Digit == DTMF_Buf[DTMF_Digits-1]) {
         // Repeated digit

            if(MinDtmfInterdigitTime != 0 && Elapsed < MinDtmfInterdigitTime) {
               bIgnoreDigit = TRUE;
            }

            LOG_ERROR(("%s: Repeated decode of \'%c\' %s, last was %d ms ago\n",
                       __FUNCTION__,Digit,
                       bIgnoreDigit ? "ignored" : "accepted",
                       Elapsed));
         }

         if(!bIgnoreDigit) {
            if(DTMF_Digits < MAX_DTMF_DIGITS) {
               DTMF_Buf[DTMF_Digits++] = Digit;
               DTMF_Buf[DTMF_Digits] = 0;
            }

            if(DTMFPrefix != NULL && !bPrefixed) {
               int Prefixed;

               if((Prefixed = CheckPrefix()) == 1) {
               // End of prefix reached
                  printf("Prefix complete\n");
                  bPrefixed = TRUE;
                  DTMF_Digits = 0;
                  if(PrefixedTone != 0) {
                  // Start dialtone
                     PlayStoredSound(PrefixedTone);
                  }
               }
               else if(Prefixed == -1) {
               // Prefix failed, search for another possible starting position
                  while(--DTMF_Digits > 0) {
                  // Remove first digit from in prefix buffer
                     memmove(DTMF_Buf,&DTMF_Buf[1],DTMF_Digits);
                     if(CheckPrefix() == 0) {
                     // another possible prefix, press on
                        break;
                     }
                  }
               }
            }
            else if(DTMF_Digits == 1) {
               // First digit
               if(Dtmf1stDigits != NULL) {
                  if(strchr(Dtmf1stDigits,Digit) == NULL) {
                     LOG_ERROR(("%s: Ignoring invalid first digit \'%c\'\n",
                                __FUNCTION__,Digit));
                     bIgnoreDigit = TRUE;
                  }
               }

               if(DTMFPrefix != NULL && PrefixedTone != 0) {
               // Cancel dialtone on first digit received
                  printf("Canceling dialtone\n");
                  CancelToneGen(PrefixedTone);
               }

               if(!bIgnoreDigit && MuteDTMF && DtmfCoverTone != 0) {
                  PlayStoredSound(DtmfCoverTone);
               }
            }
         }
      }
      LastDTMFDigit = TimeNow;

      EventHook("dtmfdecode %c",Digit == ' ' ? 'N' : Digit);
   }
   else if(Digit != ' ') {
      if(DtmfCTCSSMode == 1) {
         DPRINTF(("Port %s: Ignoring TouchTone: %c, DtmfCTCSSMode: 1,"
                  " bRxActive: %d\n",NodeName,Digit,bRxActive));
      }
      else {
         DPRINTF(("Port %s: Ignoring TouchTone: %c, DtmfCTCSSMode: 2,"
                  " bCtcssActive: %d\n",NodeName,Digit,bCtcssActive));
      }
   }
}

void Node::GenCTCSS(int16 *CTCSSBuf)
{
   int i;
   double CTCSSAngleDelta;

   CTCSSAngleDelta = 2 * M_PI * EchoCTCSSFreq / 8000.0;

   for(i = 0; i < AUDIO_BUF_SIZE; i++) {
      CTCSSBuf[i] = (int16) (CTCSSLevel * sin(CTCSSAngle));
      CTCSSAngle += CTCSSAngleDelta;
   }
}

void Node::GenCW(int16 *CwBuf)
{
   int i = 0;
   int CopyCount;
   char c;

   while(i < AUDIO_BUF_SIZE) {
      if(ElementCount == 0) {
         if(bCwToneOn) {
         // Send intra-character space
            bCwToneOn = FALSE;
            ElementCount = DitSamples;
            bElementSpace = TRUE;
            RampCount = CwRampCount;
         }
         else if(bElementSpace) {
         // next element
            bElementSpace = FALSE;
            if((c = *Element++) != 0) {
               bCwToneOn = TRUE;
               RampCount = CwRampCount;
               if(c == '-') {
                  ElementCount = DitSamples * 3;
               }
               else if(c == '.') {
                  ElementCount = DitSamples;
               }
            }
            else {
            // Next character, already did a one dit time for the element
            // space, do 2 more for character space
               ElementCount = DitSamples * 2;
            }
         }
         else {
         // character space complete, do next character
            c = *CWChar;

            if(c == 0) {
            // That's it
               CWChar = NULL;
               if(i > 0) {
                  memset(&CwBuf[i],0,(AUDIO_BUF_SIZE-i) * sizeof(int16));
                  i = AUDIO_BUF_SIZE;
               }
               break;
            }
            else if(c == ' ') {
            // Word space, 3 dit time space already sent.  Send 4 more
               ElementCount = DitSamples * 4;
               CWChar++;
            }
            else {
               for( ; ; ) {
                  if((c = *CWChar) == 0) {
                  // end of string
                     CWChar = NULL;
                     bCwIsID = FALSE;
                     break;
                  }
                  CWChar++;
                  Element = CwLookUp(c);
                  if(*Element == '-' || *Element == '.') {
                  // Start sending the next character
                     bElementSpace = TRUE;
                     break;
                  }
               }
            }
         }
      }

      if(RampCount > 0) {
      // Ramping tone level up or down
         if(bCwToneOn) {
         // Ramping up
            while(ElementCount-- > 0 && i < AUDIO_BUF_SIZE && RampCount > 0) {
               CwBuf[i] = (int16) ((ToneBuf[i] * (CwRampCount - RampCount)) / 
                                   CwRampCount);
               i++;
               RampCount--;
            }
         }
         else {
            // Ramping tone level down
            while(ElementCount-- > 0 && i < AUDIO_BUF_SIZE && RampCount > 0) {
               CwBuf[i] = (int16) ((ToneBuf[i] * RampCount--) / CwRampCount);
               i++;
            }
         }
      }
      else if(bCwToneOn) {
      // Copy full level tone
         CopyCount = min(ElementCount,AUDIO_BUF_SIZE-i);
         memcpy(&CwBuf[i],&ToneBuf[i],CopyCount*sizeof(int16));
         i += CopyCount;
         ElementCount -= CopyCount;
      }
      else {
      // No tone, just zero the buffer
         CopyCount = min(ElementCount,AUDIO_BUF_SIZE-i);
         memset(&CwBuf[i],0,CopyCount*sizeof(int16));
         i += CopyCount;
         ElementCount -= CopyCount;
      }
   }
}

/* tone generation

   .tonegen [<tone1 freq>] [<tone2 freq>] [<level>] <duration> ...

   #<id>         Play saved ToneSpec #<id>  
   !<id>         Cancel ToneSpec #<id>  
   TF1 <freq>     Set Tone 1 frequency, 0 = no tone
   EF1 <freq>     Set Tone 1 end frequency for sweeps (starts at TF1)
   TL <level>     Set Tone 1 and 2 level

   TF2 <freq>     Set Tone 2 frequency, 0 = no tone
   TL2 <level>    Set Tone 2 level
   DUR <duration> duration of tone segment in milliseconds.
   
   RPT <count>    repeat pattern from the beginning or last rpt count times,
                  -1 means repeat until cancelled.

   Each tone segment must have an duration, all other fields are optional.
   
   Tone frequencies are cleared set to zero after each segment is generated.
   Tone levels are only changed if explicitly modified.
   
   dial tone is a continuous tone having frequencies of 350 and 440 Hz at a 
   level of -13 dBm

   ringback tone is defined as comprising frequencies of 440 and 480 Hz at a 
   level of -19 dBm and a cadence of 2 seconds ON and 4 seconds OFF

   busy tone is defined as having frequency components of 480 and 620 Hz at a 
   level of -24 dBm and a cadence of half a second ON and half a second OFF

   reorder tone, also called "fast busy" tone, contains the same frequency 
   components as busy tone at a similar level but with a cadence of 0.25 of 
   a second ON and 0.25 of a second OFF
*/
void CmdToneGen(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   char *pErrPos;
   char *ToneSpec = Arg;
   int ToneID;

   do {
      if(*Arg == '#') {
      // play a stored tonespec
         if(sscanf(&Arg[1],"%d",&ToneID) != 1 || ToneID <= 0) {
            BufPrintf(p,"Error: invalid tone ID \"%s\"\r",Arg);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }

         if((ToneSpec = FindToneSpec(ToneID)) == NULL) {
            BufPrintf(p,"Error: tone #%d is not defined.\r",ToneID);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }
      }
      else if(*Arg == '!') {
      // Cancel a tonespec that's currently running
         if(sscanf(&Arg[1],"%d",&ToneID) != 1 || ToneID <= 0) {
            BufPrintf(p,"Error: invalid tone ID \"%s\"\r",Arg);
            Rcode = TBD_ERR_INVALID_ARG;
            break;
         }
         if(!CurrentNode->CancelToneGen(ToneID)) {
            BufPrintf(p,"tone #%d was not running\r",ToneID);
         }
         break;
      }

      CToneGen *pGenerator = new CToneGen();

      if(pGenerator == NULL) {
         LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
         BufPrintf(p,"Error: new failed\r");
         break;
      }

      if(pGenerator->ToneGenSetup(ToneSpec,&pErrPos)) {
      // error
         BufPrintf(p,"Error: couldn't parse \"%s\"\r",pErrPos);
         Rcode = TBD_ERR_INVALID_ARG;
         delete pGenerator;
         break;
      }
      else {
          CurrentNode->AddToneGen(pGenerator);
      }
   } while(FALSE);
}

int Node::Shutup()
{
   int Ret = FALSE;

   while(pPlayHead != NULL) {
      Ret = TRUE;
      FreeAnnoucement();
   }

   return Ret;
}

void CmdShutup(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   if(CurrentNode->Shutup()) {
      BufPrintf(p,"ok\r");
   }
   else {
      BufPrintf(p,"no announcements are playing\r");
   }
}

void CmdSayResult(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   char ErrMsg[120];
   int  bHandled = FALSE;
   int i;

// Search for a custom say result handler for this command

   for(i = 0; SayResultHandlers[i].Cmd != NULL; i++) {
      if(strcmp(SayResultHandlers[i].Cmd,LastCmd) == 0) {
      // Found it
         bHandled = SayResultHandlers[i].Handler(p,pCC1,Arg);
         break;
      }
   }

   if(!bHandled) {
      if(TTS_ExePath) {
      // We have TTS

         if(LastCmdText != NULL) {
         // Something to say, just say it
            CurrentNode->SayPhrase(LastCmdText);
         }
         else if(LastRcode == TBD_OK) {
            CurrentNode->SayPhrase("O K");
         }
         else if(LastRcode > TBD_ERR_BASE && LastRcode < TBD_LAST_ERR_CODE) {
            CurrentNode->SayPhrase(StdTbdCommandText[LastRcode - TBD_ERR_BASE - 1]);
         }
         else {
            snprintf(ErrMsg,sizeof(ErrMsg),"command failed %s",
                     Err2String(LastRcode));
            CurrentNode->SayPhrase(ErrMsg);
         }
      }
      else {
      // No TTS
         if(LastRcode == TBD_OK) {
            CurrentNode->SayPhrase("O");
            CurrentNode->SayPhrase("K");
         }
         else {

         }
      }
   }
}

int Comment(char *cp)
{
   while(*cp && isspace(*cp))
      cp++;
   if(*cp == ';' || *cp == 0 )
      return 1;
   return 0;
}

#define MAX_LINE_LEN 120
int Node::LoadCommandFile(struct config_entry *p, char *Arg)
{
   int Ret = 0;
   int LineNum = 0;
   FILE *fp = NULL;
   char  Line[MAX_LINE_LEN];
   char  DTMF_String[MAX_LINE_LEN+1];
   char  CommandString[MAX_LINE_LEN+1];
   char  *cp;
   char  *cp1;
   DTMF_Command *pDTMFCommand;
   DTMF_Command *pTail = (DTMF_Command *) &pDTMFCommands;
   char Filename[PATH_MAX];
   
   if(Arg != NULL) {
      while((pDTMFCommand = pDTMFCommands) != NULL) {
      // Clean up old commands before loading new ones
         pDTMFCommands = pDTMFCommand->Link;
         if(pDTMFCommand->DTMF_String != NULL) {
            free(pDTMFCommand->DTMF_String);
         }
         if(pDTMFCommand->Command != NULL) {
            free(pDTMFCommand->Command);
         }
         free(pDTMFCommand);
      }

      if(Arg[0] == '/') {
         strcpy(Filename,Arg);
      }
      else {
         snprintf(Filename,PATH_MAX,"%s/%s",WorkingDir,Arg);
      }

      if((fp = fopen(Filename,"r")) == NULL) {
         LOG_ERROR(("%s: Unable to open \"%s\" - %s\n",__FUNCTION__,Filename,
                    Err2String(errno)));
      }
      else {
         if(CommandFilename != NULL) {
            free(CommandFilename);
         }
         CommandFilename = strdup(Filename);
      }
   }
   
   while(fp != NULL) {
      if(fgets(Line,sizeof(Line),fp) == NULL) {
         if(ferror(fp)) {
            LOG_ERROR(("%s: read error while reading \"%s\" - %s\n",
                       __FUNCTION__,Filename,Err2String(errno)));
            Ret = errno;
         }
         break;
      }
      LineNum++;
      if((cp = strrchr(Line,'\n')) != NULL) {
         // remove terminating <lf>
         *cp = 0;
      }

      if((cp = strrchr(Line,'\r')) != NULL) {
      // remove terminating <cr> just in case some one uses an editor that
      // changes the end of line convention in *nix
         *cp = 0;
      }

      if(Comment(Line)) {
      // ignore this line if it's a comment
         continue;
      }

      if(sscanf(Line," %[^= ] = %[^\n]",DTMF_String,CommandString) != 2) {
      // line didn't convert
         LOG_WARN(("nError: Unable to parse line:\n  %s\n",Line));
         continue;
      }

   // Remove spaces from DTMF string and test for valid characters
      cp1 = cp = DTMF_String;
      while(*cp) {
         *cp = toupper(*cp);
         if((*cp >= 'A' && *cp <= 'D') || (*cp >= '0' && *cp <= '9') ||
            *cp == '#' || *cp == '*' || *cp == '?' || *cp == '+')
         {  // Valid DTMF or wildcard character, copy it

            *cp1++ = *cp++;
         }
         else if(*cp == ' ' || *cp == '\t') {
         // space or tab, ignore it
            cp++;
         }
         else {
         // error
            LOG_WARN(("Error: \"%c\" is not an DTMF character, ignoring line:\n"
                      "%s\n",*cp,Line));
         }
      }

      if(*cp) {
         continue;
      }

      *cp1 = 0;

      cp = CommandString;
   // Skip white space at beginning of the command string
      while(isspace(*cp)) {
         cp++;
      }

   // Good command, add it to the command chain

      if((pDTMFCommand = (DTMF_Command *) malloc(sizeof(DTMF_Command))) == NULL) {
         LOG_ERROR(("%s:#d: malloc failed.\n",__FUNCTION__,__LINE__));
         break;
      }

      pDTMFCommand->Link = NULL;
      pDTMFCommand->DTMF_String = strdup(DTMF_String);
      pDTMFCommand->Command = strdup(cp);
      pTail->Link = pDTMFCommand;
      pTail = pDTMFCommand;
   }

   if(fp != NULL) {
      fclose(fp);
   }

   return Ret;
}

int LoadCommandFile(struct config_entry *p, char *Arg, AfCmd Cmd, int x)
{
   int Ret = -1;

   switch(Cmd) {
      case AF_SET_VAR:
         if(ConfigPass > 0) {
            Ret = CurrentNode->LoadCommandFile(p,Arg);
         }
         break;

      case AF_SET_DEFAULT:
         break;

      case AF_DISPLAY_VAR:
         if(CurrentNode->CommandFilename == NULL) {
            snprintf(Arg,x,"(not set)");
         }
         else {
            strncpy(Arg,CurrentNode->CommandFilename,x);
         }
         break;
   }
   
   return Ret;
}

void CmdSay(ClientInfo *p,ConfClient *pCC,char *Arg)
{
   char *cp = Arg;
   int ReplyPort = 0;

   if(pCC->bCmdLine) {
   // look for -c, if found then send TBD_SAY_COMPLETE on completion
      while(*cp == ' ') {
         cp++;
      }
      if(*cp == '-' && cp[1] == 'c') {
         ReplyPort = pCC->HisAdr.PORT;
         cp += 2;
      }
      while(*cp == ' ') {
         cp++;
      }
   }
   
   if(CurrentNode != NULL) {
      CurrentNode->SayPhrase(cp,ReplyPort);
   }
}

void CmdSweep(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   int StartFreq;
   int EndFreq;
   int SweepTime;
   int Sweeps;
   int Level = 0;
   int Converted;

   if(CurrentNode != NULL) {
      Converted = sscanf(Arg,"%d %d %d %d %d",
                         &StartFreq,&EndFreq,&SweepTime,&Sweeps,&Level);
      switch(Converted) {
         case 1:
         // Just start frequency, default the rest
            EndFreq = StartFreq;
         case 2:
            SweepTime = 15;
         case 3:
            Sweeps = 1;
         case 4:
            Level = 30000;
            break;

         default:
            break;
      }

      if(Level != 0) {
         char Line[80];

         snprintf(Line,sizeof(Line),"TF1 %d EF1 %d TL %d DUR %d RPT %d",
                  StartFreq,EndFreq,Level,SweepTime*1000,Sweeps);

         char *pErrPos;
         CToneGen *pGenerator = new CToneGen();

         if(pGenerator->ToneGenSetup(Line,&pErrPos)) {
         // error
            BufPrintf(p,"Error: couldn't parse \"%s\"\r",pErrPos);
            Rcode = TBD_ERR_INVALID_ARG;
            delete pGenerator;
         }
         else {
            CurrentNode->AddToneGen(pGenerator);
         }
      }
      else {
         BufPrintf(p,"Usage: .sweep <start freq> <end freq> <sweep time> "
                   "<sweeps> <sweep level>\r");
      }
   }
}

void CmdDtmfDecode(ClientInfo *p,ConfClient *pCC1,char *Arg)
{
   CurrentNode->DecodeDTMF(Arg);
}

void Node::AddToneGen(CToneGen *pGenerator)
{
   CToneGen *p = NULL;

   if(pGenerator->ID != 0) {
   // Only allow one tone generator of the same id to be running at a time
   // prevents arcade like sounds on weak signals among other annoying things!
      p = ToneGenerators;
      while(p != NULL) {
         if(p->ID == pGenerator->ID) {
         // Duplicate generator
            delete pGenerator;
            break;
         }
         p = p->Link;
      }
   }

   if(p == NULL) {
      pGenerator->Link = ToneGenerators;
      ToneGenerators = pGenerator;
      CPort *pPort = dynamic_cast<CPort *> (this);
      if(pPort != NULL && !pPort->pAudioC->bWriteWait) {
         pPort->WriteAudioOut(pPort->pAudioC);
      }
   }
}

int Node::CancelToneGen(int id)
{
   int Ret = FALSE;
   CToneGen *pToneGen;
   CToneGen *pLastGen;
   CToneGen *pNextGen;

   pToneGen = ToneGenerators;
   pLastGen = NULL;
   while(pToneGen != NULL) {
      pNextGen = pToneGen->Link;
      if(pToneGen->ID == id) {
      // Delete this generator
         Ret = TRUE;
         if(pLastGen == NULL) {
            ToneGenerators = pNextGen;
         }
         else {
            pLastGen->Link = pNextGen;
         }
         delete pToneGen;
         break;
      }
      else {
         pLastGen = pToneGen;
      }
      pToneGen = pNextGen;
   }
   return Ret;
}

// Send CTone or RxTimeout tone to all transmitters that this receiver
// is connected to
void Node::SendCtone(int bDelayed)
{
   int ToneID = bRxTimeout ? RxTimeoutTone : Ctone;
   int bDelayedCtone = FALSE;

   if(ToneID != 0 && CtoneMap != NULL) do {
   // We have a tone ID and need to send it to least one port
      char *ToneSpec = FindToneSpec(ToneID);

      if(ToneSpec == NULL) {
         LOG_ERROR(("%s: Tone ID #%d is not defined, disabling.\n",
                    __FUNCTION__,ToneID));
         if(bRxTimeout) {
            RxTimeoutTone = 0;
         }
         else {
            Ctone = 0;
         }
         break;
      }

      LinkedNode *pLink = pTransmitters;
      while(pLink != NULL) {
      // This port is listening to our Rx
         if(!pLink->pTx->bTxTimeOut) {
            if(!bRxTimeout && CtoneDelay != 0 && !bDelayed) {
            // Send it later
               bDelayedCtone = TRUE;
               break;   // No need to look further
            }

            if(!bRxTimeout && pLink->pTx->CtoneResetsTimeout) {
            // Reset the transmitter's timeout timer
               pLink->pTx->TxKeyTime = TimeNow.tv_sec;
            }

            if(pLink->Flags & LNK_FLG_CTONE) {
            // CTone enabled to this transmitter
               CToneGen *pToneGen = new CToneGen;
               if(pToneGen != NULL) {
                  if(pToneGen->ToneGenSetup(ToneSpec,NULL)) {
                      delete pToneGen;
                  }
                  else {
                      pLink->pTx->AddToneGen(pToneGen);
                  }
               }
               else {
                  LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
               }
            }
         }
         pLink = pLink->pNextTx;
      }
      
      if(bDelayedCtone != 0) {
      // Generating a delayed courtesy tone, start the timer
         CtoneTimer = TimeNow;
      }
   } while(FALSE);
}

// return TRUE on error
int Node::PlayStoredSound(int ID)
{
   int Ret = 0;
   do {
      if(ID == 0) {
         LOG_ERROR(("%s: Called with ID #%d.\n",__FUNCTION__));
         break;
      }
      char *ToneSpec = FindToneSpec(ID);

      if(ToneSpec == NULL) {
         LOG_ERROR(("%s: ID #%d is not defined.\n",__FUNCTION__,ID));
         Ret = TRUE;
         break;
      }
      CToneGen *pGenerator = new CToneGen();

      if(pGenerator != NULL) {
         if(pGenerator->ToneGenSetup(ToneSpec,NULL)) {
             delete pGenerator;
         }
         else {
             AddToneGen(pGenerator);
         }
      }
      else {
         LOG_ERROR(("%s#%d: new failed\n",__FUNCTION__,__LINE__));
         Ret = TRUE;
      }
   } while(FALSE);

   return Ret;
}

int Node::InitToneBuf(struct config_entry *p, char *Arg)
{
   double ToneFreq;
   double Cycles;
   double BufDuration;
   int i;
   int Ret = 1;   // assume the worse
   void *var_ptr;       // pointer to variable

   if(Arg == NULL || p == NULL) {
      Ret = 0;
   }
   else {
      var_ptr = ConfigGetVarPtr(p);
      if(sscanf(Arg,p->var_type,var_ptr) == 1) {
      // fudge the ID tone frequency so an integer number of cycles fit
      // into our tone buffer
      
         BufDuration = AUDIO_BUF_SIZE/8000.0;
         Cycles = floor(CWToneFreq * BufDuration);
         ToneFreq = Cycles / BufDuration;
      
         for(i = 0; i < AUDIO_BUF_SIZE; i++) {
            ToneBuf[i] = (int16) (CWToneLevel * sin(i*Cycles*2*M_PI / AUDIO_BUF_SIZE));
         }
      
      // Calculate number of cycles of tone in a dit.
      //
      // The word PARIS was chosen as the standard length for CW code speed. 
      // Each dit counts for one count, each dah counts for three counts, 
      // intra-character spacing is one count, inter-character spacing is three 
      // counts and inter-word spacing is seven counts, so the word PARIS is 
      // exactly 50 counts.
      // So 1 wpm = 50 counts/minute or 50/60 counts/second.
      // or Samples/dit = (60/50 secs/dit) * (AUDIO_BUF_SIZE samples) / 
      //                  (BufDuration secs);
      
         DitSamples = (int) ((60.0 * AUDIO_BUF_SIZE) / 
                             (50.0 * BufDuration * CWSpeed));  
      
      // Calculate attach and decay time for tone envelope
      // 10% of dit time ramping up, 80% @ full tone level, 10% ramping down
      
         CwRampCount = DitSamples * 10 / 100;
         Ret = 0;
      }
   }

   return Ret;
}

int Node::Enabled()
{
   return TRUE;
}

void Node::CheckDtmfTimeouts(void)
{
   if(DTMF_Digits > 0) {
      if(TimeLapse(&LastDTMFDigit) > DTMFCommandTimeout) {
         printf("DTMFCommandTimeout\n");
         if(DTMFPrefix == NULL || bPrefixed) {
            DecodeDTMF(DTMF_Buf);
         }
         DTMF_Digits = 0;
         if(MuteDTMF && DtmfCoverTone != 0) {
            CancelToneGen(DtmfCoverTone);
         }
         LastDTMFDigit.tv_sec = 0;  // force a prefix timeout on command timeout
      }
   }

   if(DTMF_Digits == 0 && bPrefixed) {
      if(TimeLapse(&LastDTMFDigit) > PrefixTimeout) {
         printf("PrefixTimeout\n");
         if(PrefixedTone != 0) {
            CancelToneGen(PrefixedTone);
         }
         bPrefixed = 0;
      }
   }
}

int InitToneBuf(struct config_entry *p, char *Arg, AfCmd Cmd, int x)
{
   int Ret = 0;
   CPort *pPort;

   switch(Cmd) {
      case AF_SET_VAR:
         pPort = dynamic_cast<CPort *> (CurrentNode);
         if(pPort != NULL) {
            Ret = pPort->InitToneBuf(p,Arg);
         }
         break;

      case AF_SET_DEFAULT:
         break;

      case AF_DISPLAY_VAR:
         break;
   }

   return Ret;
}

int SayConnectResult(ClientInfo *p,ConfClient *pCC,char *Arg)
{
   int Ret = FALSE;

   if(LastRcode == TBD_OK && CurrentNode != NULL) {
      if(StationDisconnected != NULL) {
         CurrentNode->SayPhrase("connecting to ");
         Ret = TRUE;
      }
   }

   return Ret;
}

int SayDisconnectResult(ClientInfo *p,ConfClient *pCC,char *Arg)
{
   int Ret = FALSE;
   
   if(LastRcode == TBD_OK && CurrentNode != NULL) {
      if(StationDisconnected != NULL) {
         CurrentNode->SayCall(StationDisconnected);
         CurrentNode->SayPhrase("disconnected");
         Ret = TRUE;
      }
   }

   return Ret;
}


